; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *                     Global Data / Setup Data                        *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************
.386p
.model flat
include gb.inc
.code

Ext_SGS equ     ('g')+('a'*256)+('0'*65536)

; ***********************************************************************
; *                                                                     *
; *                     Check for game save/load                        *
; *                                                                     *
; ***********************************************************************

CheckLoadSave Proc
        push    eax ebx
        mov     eax,[SaveKey]
        call    KTRelease
        test    al,128
        jz      NoSave
        mov     edx,offset SaveTo
        call    Slot
        cmp     [StatusMask],0
        je      Abandon
        call    SaveGameStatus
        mov     ebx,320*192                     ; display information
        call    TxtEOL
        mov     al,14
        call    TxtColour
        mov     edx,offset Saved
        call    TxtString
NoSave: mov     eax,[LoadKey]
        call    KTRelease
        test    al,128
        jz      NoLoad
        mov     edx,offset LoadTo
        call    Slot
        cmp     [StatusMask],0
        je      Abandon
        call    LoadGameStatus
        mov     ebx,320*192                     ; display information
        call    TxtEOL
        mov     al,14
        mov     edx,offset Loaded
        call    TxtString
NoLoad: pop     ebx eax
        ret
Abandon:mov     ebx,320*192
        call    TxtEOL
        jmp     NoLoad
.data
Saved   db      'Game Saved',0
Loaded  db      'Game Loaded',0
SaveTo  db      'Save to which slot (0-7)',0
LoadTo  db      'Load from which slot (0-7)',0
.code
KTRelease:
        mov     ebx,eax
        mov     al,[byte ptr KeyTable+eax]
        test    al,128
        jz      KTRExit
KTRWait:mov     al,[byte ptr KeyTable+ebx]
        test    al,128
        jnz     KTRWait
        mov     al,128
KTRExit:ret

Slot:   push    eax ebx edx
        mov     [StatusMask],0
        mov     ebx,320*192
        call    TxtEOL
        mov     al,9
        call    TxtColour
        pop     edx
        call    TxtString
        mov     ebx,offset KeyList
SWait:  call    GetKeyPressed
        cmp     al,-1
        je      SWait
        cmp     al,8
        je      SlotExit
        and     al,7
        or      al,48
        mov     [byte ptr StatusMask+2],al
SlotExit:
        pop     ebx eax
        ret
KeyList db 0bh,2,3,4,5,6,7,8,1Ch,255
EndP CheckLoadSave

; ***********************************************************************
; *                                                                     *
; *         Find key pressed in list ^EBX, return # in AL or 255        *
; *                                                                     *
; ***********************************************************************

GetKeyPressed Proc
        push    ebx ecx
        xor     ecx,ecx
        xor     eax,eax
        mov     al,255
GKP1:   mov     cl,[ebx]
        inc     ebx
        inc     al
        cmp     cl,255
        je      GKPFail
        mov     cl,[byte ptr KeyTable+ecx]
        test    cl,128
        jz      GKP1
        jmp     GKPExit
GKPFail:mov     al,255
GKPExit:pop     ecx ebx
        ret
EndP GetKeyPressed

; ***********************************************************************
; *                                                                     *
; *                      Load the game status                           *
; *                                                                     *
; ***********************************************************************

LoadGameStatus Proc
        push    edi esi edx ecx ebx eax
        mov     eax,EXT_SGS                     ; load the file
        or      eax,[StatusMask]                ; if it doesn't exist
        call    FileNameExt                     ; do nothing
        mov     esi,offset SavedData
        call    LoadFile
        cmp     ax,-1
        je      LGSExit

        call    Reset                           ; reset the CPU
        call    ResetData                       ; reset the BSS data
        call    ResetMapping                    ; reset the mapping
        call    ResetMappingData                ; and the masks
        call    ClearMemory                     ; clear memory
        mov     eax,EXT_SGS                     ; load the file
        or      eax,[StatusMask]
        call    FileNameExt
        mov     esi,offset SavedData
        call    LoadFile
        cmp     ax,-1
        je      LGSExit
        mov     esi,offset CPUData              ; loaded ok. copy CPU
        mov     edi,offset CPURStart            ; status information back.
LGS1:   mov     al,[esi]
        mov     [edi],al
        inc     esi
        inc     edi
        cmp     edi,offset CPUREnd
        jne     LGS1

        xor     eax,eax                         ; load in the current rom page
        mov     al,[CurrROMPage]
        shl     eax,14                          ; multiply by 16384
        add     eax,offset ROMImage             ; eax points to ROM start
        call    MapROMPage                      ; redo the mapping
        mov     al,[CurrRAMPage]                ; select current RAM Page
        cmp     al,255                          ; fix if never selected by
        jne     NotSel                          ; code its zero
        xor     al,al
NotSel: call    MapRAMPage

LGSExit:pop     eax ebx ecx edx esi edi
        ret
EndP LoadGameStatus

; ***********************************************************************
; *                                                                     *
; *                      Save the game status                           *
; *                                                                     *
; ***********************************************************************

SaveGameStatus:
        push    edi esi edx ecx ebx eax
        mov     esi,offset CPUData              ; copy CPU Status -> Save Area
        mov     edi,offset CPURStart
Copy1:  mov     eax,[edi]
        mov     [esi],eax
        add     esi,4
        add     edi,4
        cmp     edi,offset CPUREnd
        jb      Copy1
        mov     eax,EXT_SGS                     ; write it to the disk.
        or      eax,[StatusMask]
        call    FileNameExt
        mov     ecx,SavedDataEnd-SavedData
        mov     eax,offset SavedData
        call    SaveFile
        pop     eax ebx ecx edx esi edi
        ret

; ***********************************************************************
; *                                                                     *
; *                          Initialise Data                            *
; *                                                                     *
; ***********************************************************************

ResetData Proc
        mov     [ROMMask],0
        mov     [CurrROMPage],255
        mov     [CurrRAMPage],255
        mov     [OldPi0To3],255
        mov     [NewFF00Value],255
        call    SerialReset
        ret
EndP ResetData

.data

; ***********************************************************************
; *                                                                     *
; *             Global Data belonging to the emulation                  *
; *                                                                     *
; ***********************************************************************

align dword
VideoRam        dd      0                       ; offset to VGA memory
PaintVector     dd      CopyLogicalPhysical     ; screen copy routine
                                                ; (CopyLogicalPhysical/Large)
StatusMask      dd      0                       ; save load status or
DisplayCounter  db      0                       ; display skip counter
KeyboardByte    db      0                       ; keyboard byte
TimerInterrupt  db      0                       ; timer interrupt flag
FileName        dd      0                       ; pointer to filename

BMStart         dw      0                       ; Benchmarking information
BMEnd           dw      0
BMFrame         dw      0

AutoOnA         db      1                       ; autofire on flags
AutoOnB         db      1

VGBScheme       dd      0                       ; colour scheme data ?
                dd      15 dup (0)

Fixed           db      0                       ; cartridge fixed ?
             
; ***********************************************************************
; *                                                                     *
; *                     Global Data the user can change                 *
; *                                                                     *
; ***********************************************************************

align dword

UserDataStart:

KeyRight        dd      4Dh             ; $ Right Key
KeyLeft         dd      4Bh             ; $ Left Key
KeyUp           dd      48h             ; $ Up Key
KeyDown         dd      50h             ; $ Down Key
KeyA            dd      1Dh             ; $ A Key
KeyB            dd      38h             ; $ B Key
KeySelect       dd      0Fh             ; $ Select Key
KeyStart        dd      1Ch             ; $ Start Key

AutoFireA       db      0               ; #0-8 Auto rate A
AutoFireB       db      0               ; #0-8 Auto rate B

FastSprites     db      0               ; b Fast Sprites

GBBackground    dd      0FFFFFFh        ; @ Back colour
GBWindow        dd      0FFFFFFh        ; @ Window colour
GBSprite        dd      0FFFFFFh        ; @ Sprite colour
                                                                
Background      db      1               ; #0-15 Border Colour
DisplayRate     db      3               ; #1-15 Frame Skip
BigScreen       db      0               ; b Large screen

SoundOn         db      1               ; b Sound Enabled
SyncOn          db      1               ; b CPU Sync On
FrameRate       dw      3               ; #1-8 Frames/Tick
ExtraRate       db      4               ; #1-10 Sneak ticks

_WindowFudge    db      0               ; b Fudge window (removed)
NoInt60         db      0               ; b No Int 60h
FMPort          dw      0388h           ; #1-999 Adlib Port
Mode10Cycles    dw      84              ; #20-500 CPU cyc 10
Mode11Cycles    dw      168             ; #20-500 CPU cyc 11
Mode00Cycles    dw      205             ; #20-500 CPU cyc 00

NoMLInterrupt   db      0               ; b No Rec. Int

EnableCheat1    db      0               ; b Enable Cheat 1
Cheat1          dd      0               ; c Genie Cheat 1
EnableCheat2    db      0               ; b Enable Cheat 2
Cheat2          dd      0               ; c Genie Cheat 2
EnableCheat3    db      0               ; b Enable Cheat 3
Cheat3          dd      0               ; c Genie Cheat 3
EnableCheat4    db      0               ; b Enable Cheat 4
Cheat4          dd      0               ; c Genie Cheat 4

AutoCtrlA       dd      3Bh             ; $ Key ctl auto A
AutoCtrlB       dd      3Ch             ; $ Key ctl auto B

JoystickPort    dw      0201h           ; #0-999 Joystick Port

_resv0          db      0               ; x Enable Colour ?

_resv1          dd      0               ; x Colour Scheme
_resv2          dd      0               ; x Colour Scheme
_resv3          dd      0               ; x Colour Scheme
_resv4          dd      0               ; x Colour Scheme

DebugKey        dd      058h            ; $ Debugger Key

HorzLineCount   db      154             ; #145-255 Horiz Scan Lines

Hack            db      0               ; #0-255 Hack Number
ResetKey        dd      0Eh             ; $ Reset Key
DisplayKey      dd      35h             ; $ Display Key
PauseKey        dd      19h             ; $ Pause Key
SaveKey         dd      41h             ; $ Save Game Key
LoadKey         dd      3Fh             ; $ Load Game Key
KeyPaint        dd      2Eh             ; $ Painter Key
VBIWhenOff      db      0               ; b VBI when off
Mode3Vol        db      6               ; #0-15 Mode 3 Vol
OrInt           db      0               ; b Or IntFlag
UseFixes        db      1               ; b Fix ROMs
EarlyVBI        db      0               ; b Early VBLInt
KillMouse       db      0               ; b No mouse check

UserDataEnd:

.data?

; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *                  Uninitialised large data spaces                    *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************

align dword

; ***********************************************************************
;     Data belonging to the emulation - saved - will be a big file !
; ***********************************************************************
SavedData:

CPUData        db 128 dup (?)           ; copy of the CPU Registers.
               db 128 dup (?)           ; for expansion
SIOBuf         db      ?                ; the serial I/O stuff.
SIOFlag        db      ?
SIOCount       db      ?

ROMMask db      ?                       ; mask value for ROM
CurrROMPage db  ?                       ; current ROM Page
CurrRAMPage db  ?                       ; current RAM Page
OldPI0to3 db    ?                       ; old value of PI0..PI3
NewFF00Value db ?                       ; value to go into $FF00

align dword

CartridgeRAM:                           ; first 32k of CPU memory (not
                                        ; used) for MBC RAM
Memory:        dd 16384 dup (?)         ; CPU Memory (64k)
               dd 32 dup (?)            ; filler

SavedDataEnd:
; ***********************************************************************
;                               End of saved data
; ***********************************************************************

               dw 16384 dup (?)

Bitmap:        dd 32768 dup (?)         ; background bitmap

DisplayImage:  dd 32768 dup (?)         ; Logical Screen

ROMImage:      dd 32768 dup (?)         ; ROM Image space (each is 128k)
               dd 32768 dup (?)         ; should be enough. Had problems
               dd 32768 dup (?)         ; allocating this much BSS.
               dd 32768 dup (?)
               dd 32768 dup (?)
               dd 32768 dup (?)
               dd 32768 dup (?)
               dd 32768 dup (?)

end

