; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *                   GUI Code - GUI Object Display                     *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************

.386p 
.model flat
.code 
include Gb.Inc

.data
obXPos          equ     0                       ; +0 = X Position
obYPos          equ     2                       ; +2 = Y Position
obWidth         equ     4                       ; +4 = Width
obHeight        equ     6                       ; +6 = Height
obBackColour    equ     8                       ; +8 = Background Colour
obBorderColour  equ     10                      ; +10 = Border Colour
obSelectColour  equ     12                      ; +12 = Background Col (selected)
obSelected      equ     14                      ; +14 = Selected flag.
obTextColour    equ     16                      ; +16 = Text Colour
obDataType      equ     18                      ; +18 = Format (0=text,1=bool,
                                                ;       2=hex,3=dec,4=button)
obFormat        equ     20                      ; +20 = bit 0 = vcent,bit 1 = hcent
obLength        equ     22                      ; +22 = length (text,0=fixed)
obData          equ     24                      ; +24 = Data (pointer or long)
obIdentifier    equ     28                      ; +28 = Identifier

DemoObj         dw      64,10,168,32            ; position
                dw      6,15,1,0,14             ; colours
                dw      0,3,3                   ; type & format
                dd      HelloWorld
                dw      4                       ; identifier.

HelloWorld      db      'Hello, world !',0
.code

; ***********************************************************************
; *                                                                     *
; *                   Draw GUI Object pointed to by EBX                 *
; *                                                                     *
; ***********************************************************************
                                                
GUIDrawObject Proc
        push    eax ebx ecx edx esi edi

        mov     ebx,offset DemoObj

        xor     ecx,ecx                         ; work out where it is.
        xor     eax,eax                         ; in physical memory
        mov     ax,[ebx+obYPos]
        mov     cx,ax
        shl     eax,2
        add     eax,ecx
        shl     eax,6
        mov     cx,[ebx+obXPos]
        add     eax,ecx
        add     eax,[VideoRAM]
        mov     edi,eax                         ; video pointer in EDI

        push    edi                             ; draw the background
        mov     ch,[ebx+obHeight]               ; ch = y counter
        mov     ah,[ebx+obBorderColour]         ; ah = border colour
GUID1:  push    edi
        mov     si,[ebx+obWidth]                ; si = width
        mov     al,[ebx+obBackColour]           ; al = back or select ?
        mov     dl,[ebx+obSelected]
        cmp     dl,0
        jz      GUID5
        mov     al,[ebx+obSelectColour]
GUID5:  cmp     ch,1                            ; top/bottom edge ?
        je      GUID3
        cmp     ch,[ebx+obHeight]
        jne     GUID2
GUID3:  mov     al,ah
GUID2:  mov     [edi],al
        inc     edi
        dec     si
        jnz     GUID2
        mov     [edi-1],ah
        pop     edi
        mov     [edi],ah
        add     edi,320
        dec     ch
        jnz     GUID1
        pop     edi

        mov     al,[ebx+obTextColour]           ; set the text colour
        call    TxtColour
        push    ebx edi
        mov     edx,[ebx+obData]                ; edx points to data
        mov     al,[ebx+obDataType]
        call    ConvertFormat
        mov     ah,[ebx+obFormat]               ; check format
        test    ah,1
        jz      NoVCent
        xor     ecx,ecx
        mov     cx,[ebx+obHeight]               ; cx = (Height-8)/2
        sub     cx,8
        jns     NotNeg1
        xor     cx,cx
NotNeg1:shr     cx,1
        shl     ecx,6
        add     edi,ecx
        shl     ecx,2
        add     edi,ecx                         ; add to text position
NoVCent:test    ah,2                            ; check format (h centre)
        jz      NoHCent
        call    StrLen                          ; al = strlen(edx)
        movzx   eax,al                          ; make it 32 bit
        shl     eax,2                           ; multiply by four
        xor     ecx,ecx
        mov     cx,[ebx+obWidth]                ; ecx = width/2
        shr     ecx,1
        sub     ecx,eax                         ; work out offset
NotNeg2:add     edi,ecx                         ; add to video pointer
NoHCent:mov     ebx,edi
        sub     ebx,[VideoRAM]                  ; ebx points to the video
        mov     [NoBgr],1
        call    TxtString                       ; draw the string
        mov     [NoBgr],0
        pop     edi ebx

        pop     edi esi edx ecx ebx eax
        ret
EndP GUIDrawObject

; ***********************************************************************
; *                                                                     *
; *                  Return length of string EDX in AL                  *
; *                                                                     *
; ***********************************************************************

StrLen Proc
        push    ecx
        xor     eax,eax
SL1:    mov     cl,[edx+eax]
        cmp     cl,0
        jz      SL2
        inc     eax
        jmp     SL1
SL2:    pop     ecx
        ret
EndP StrLen

; ***********************************************************************
; *                                                                     *
; *              Convert data in EDX to display format AL               *
; *                                                                     *
; ***********************************************************************

ConvertFormat Proc
        cmp     al,1
        jz      CFBoolean
        cmp     al,2
        jz      CFHex
        cmp     al,3
        jz      CFDec
        ret
CFHex:  mov     ecx,edx
        mov     edx,offset CFBuffer+1
        mov     [CFBuffer],'$'
        call    CFHex2
        xor     cl,cl
        mov     [edx],cl
        mov     edx,offset CFBuffer
        ret
CFDec:  mov     ecx,edx
        mov     edx,offset CFBuffer+1
        mov     [CFBuffer],'$'
        call    CFDec2
        xor     cl,cl
        mov     [edx],cl
        mov     edx,offset CFBuffer
CFDec2: cmp     ecx,10
        jb      CFHex3
        push    ecx
        call    DivECX10
        call    CFDec2
        pop     ecx
        jmp     CFHex3

DivECX10:
        push    eax ebx edx
        xor     edx,edx
        mov     eax,ecx
        mov     ecx,10
        idiv    ecx
        mov     ecx,eax
        pop     edx ebx eax
        ret

CFHex2: cmp     ecx,16
        jb      CFHex3
        push    ecx
        shr     ecx,4
        call    CFHex2
        pop     ecx
CFHex3: and     ecx,15
        add     ecx,48
        cmp     ecx,58
        jb      CFHex4
        add     ecx,7
CFHex4: mov     [edx],cl
        inc     edx
        ret

CFBoolean:
        cmp     edx,0
        jz      CFFalse
        mov     edx,offset CFT1
        ret
CFFalse:mov     edx,offset CFT2
        ret
.data
CFBuffer db     32 dup (0)
CFT1:   db      'True',0
CFT2:   db      'False',0
.code
EndP ConvertFormat


end

