/************************************************************************/
/*                                                                      */
/*						Instruction Generation							*/
/*                                                                      */
/************************************************************************

Generates : Groups 1,2,3,4

*/

#include "cpu.h"

/************************************************************************/
/*                                                                      */
/*				Group 1 (register to register transfers)				*/
/*                                                                      */
/************************************************************************/

void Group1(void)
{
int Source,Target;
char Descr[16];
for (Target = 0;Target < 8;Target++)	/* Where its going */
	for (Source = 0;Source<8;Source++)	/* Where its coming from */
		{
		if (Source != REG_HL ||			/* no ld (hl),(hl) */
						Target != REG_HL)
			{
			strcpy(Descr,"ld ");		/* Build the description */
			strcat(Descr,GetRegName(Target));
			strcat(Descr,",");
			strcat(Descr,GetRegName(Source));
			Define(0x40+Target*8+Source,/* Define the instruction */
					(Source == REG_HL || Target == REG_HL) ? 8 : 4,
					Descr);
			if (Source != Target)		/* No code if they are the same */
				{
				GetReg(Source,"al");	/* GetReg() and PutReg() handle */
				PutReg(Target,"al");	/* (HL) automatically ! */
				}
			else
				{                       /* If were tracing the ld c,c (on) */
				#ifdef TRACING			/* and ld b,b (off) control listing */
				if (Source < 2)
					printf("    mov [TraceFlag],%d\n",Source);
				#endif
				}
			EndDefine();
			}
		}
}

/************************************************************************/
/*                                                                      */
/*				Group 2 - Arithmetic register operations				*/
/*                                                                      */
/************************************************************************/

void Group2(void)
{
int Op,Source;
char Descr[16];
for (Op = 0;Op < 8;Op++)				/* Work through operations */
	for (Source = 0;Source<8;Source++)	/* Various second operands */
		{								/* first is Acc */
		sprintf(Descr,"%s a,%s",GetOpName(Op),GetRegName(Source));
		Define(0x80+Op*8+Source,(Source == REG_HL ? 8 : 4),Descr);
		GetReg(Source,"al");			/* Get the value */
		Operation(Op);					/* Do the operation */
		EndDefine();
		}
}

/************************************************************************/
/*                                                                      */
/*					Group 3 - immediate operations						*/
/*                                                                      */
/************************************************************************/

void Group3(void)
{
int Op,Reg;
char Descr[32];
for (Reg = 0;Reg < 8;Reg++)				/* Move immediate */
	{
	sprintf(Descr,"ld %s,#nn",GetRegName(Reg));
	Define(0x06+Reg*8,(Reg == REG_HL ? 12 : 8),Descr);
	Fetch("al");						/* Get the value */
	PutReg(Reg,"al");					/* store it (HL) is automatic */
	EndDefine();
	}
for (Op = 0;Op < 8;Op++)				/* Arithmetic immediate */
	{
	sprintf(Descr,"%s a,#nn",GetOpName(Op));
	Define(0xC6+Op*8,8,Descr);
	Fetch("al");  						/* Get immediate value */
	Operation(Op);						/* Do the operation */
	EndDefine();
	}
}

/************************************************************************/
/*                                                                      */
/*			Group 4 - Increment and Decrement Registers					*/
/*                                                                      */
/************************************************************************/

void Group4(void)
{
int Reg,Dir;
char Descr[16];
for (Reg = 0;Reg < 8;Reg++)				/* Work through 8 regs (inc (HL)) */
	for (Dir = 0;Dir < 2;Dir++)			/* both INC & DEC */
		{
		sprintf(Descr,"%s %s",(Dir == 0 ? "inc" : "dec"),GetRegName(Reg));
		Define(0x04+Dir+Reg*8,(Reg == REG_HL ? 12 : 4),Descr);
		GetReg(Reg,"al");				/* Get value */
		Operation(Dir == 0 ? O_INC8 :	/* Do the operation */
								O_DEC8);
		PutReg(Reg,"al");				/* Write it back */
		EndDefine();
		}
}