; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *                    Background Bitmap Functions                      *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************

.386p 
.model flat 
.code 
include Gb.Inc

.data
PaintCursor     db      0               ; currently selected paint cursor
Changed         db      0
OldTimer        dw      0               ; old timer
.code

Ext_CSC equ     ('C')+('9'*256)+('7'*65536)

; ***********************************************************************
; *                                                                     *
; *                    Check for Paint Invocation                       *
; *                                                                     *
; ***********************************************************************

CheckPaint  Proc
        push    eax
        mov     eax,[KeyPaint]
        mov     al,[byte ptr KeyTable+eax]
        test    al,128
        jz      NoPaint
        call    Painter
NoPaint:pop     eax
        ret
EndP CheckPaint

; ***********************************************************************
; *                                                                     *
; *                    Invoke the painter                               *
; *                                                                     *
; ***********************************************************************

Painter Proc
        push    eax ebx ecx edx
        mov     [Changed],0                     ; clear changed flag
        mov     eax,[dword ptr VGBScheme]       ; colour scheme loaded ?
        add     eax,[dword ptr VGBScheme+4]
        add     eax,[dword ptr VGBScheme+8]
        add     eax,[dword ptr VGBScheme+12]
        jnz     PaintSchemeLoaded
        call    PaintDefault
PaintSchemeLoaded:
        mov     [PaintCursor],0
        jmp     PaintMain
PaintDown:
        add     [PaintCursor],2
PaintUp:dec     [PaintCursor]
PaintMain:
        and     [PaintCursor],15
        mov     ebx,320*192                     ; display information
        call    TxtEOL
        mov     al,14
        call    TxtColour
        xor     eax,eax
        mov     al,[PaintCursor]
        shr     eax,2
        mov     edx,[DescTable+eax*4]
        call    TxtString
        mov     al,' '
        call    TxtCharacter
        mov     al,[PaintCursor]
        and     al,3
        add     al,'0'
        call    TxtCharacter
PaintLoop:                                      ; wait for key to be pressed
        call    FlashPaint
        mov     ebx,offset PainterKeys
        call    GetKeyPressed
        cmp     al,-1
        je      PaintLoop
        push    eax                             ; set the colour back
        call    LoadColourScheme
PLRel:  mov     ebx,offset PainterKeys
        call    GetKeyPressed
        cmp     al,-1
        jne     PLRel
        pop     eax                             ; deal with the keys
        cmp     al,2
        je      PaintUp
        jb      ChangeColour
        cmp     al,3
        je      PaintDown
        cmp     al,4
        jne     PaintMain
        call    ClearScreen
        call    [PaintVector]
        cmp     [Changed],0             ; if changed, update colour scheme
        jz      NoChange
        mov     eax,Ext_CSC
        call    FileNameExt
        mov     ecx,64
        mov     eax,offset VGBScheme
        call    SaveFile
NoChange:
        pop     edx ecx ebx eax
        ret

ChangeColour:
        mov     [Changed],1
        push    ax
        xor     eax,eax                         ; ecx points to the table
        mov     al,[PaintCursor]                ; entry
        lea     ecx,[DWord Ptr VGBScheme+eax*4]
        mov     ebx,0                           ; search the table
Search: mov     eax,[DWord Ptr ColourTable+ebx*4] ; check colour
        inc     ebx
        cmp     eax,-1                          ; until end of table
        je      EndCC
        cmp     eax,[ecx]                       ; or matching
        jne     Search
EndCC:  pop     ax
        cmp     ax,0
        jz      Back
        mov     eax,[DWord Ptr ColourTable+ebx*4] ; get the next one
        cmp     eax,-1                          ; if end back to start
        jne     Update
        mov     eax,[DWord Ptr ColourTable]

Update: mov     [ecx],eax
        call    LoadColourScheme
        jmp     PaintMain

Back:   mov     eax,[DWord Ptr ColourTable-8+ebx*4]
        cmp     eax,-1
        jne     Update
        mov     eax,[DWord Ptr LastColour]
        jmp     Update

.data
PainterKeys     db      04Bh,04Dh,048h,050h,1Ch,-1   ; Left/Right/Up/Down/Ret

DescTable       dd      ms1,ms2,ms3,ms4         ; messages
ms1     db      'Background',0
ms2     db      'Window',0
ms3     db      'Sprite A',0
ms4     db      'Sprite B',0

ColourTable:
        dd      0000000h
        dd      00000FFh
        dd      000FF00h
        dd      055FFFFh
        dd      0FF0000h
        dd      0FF55FFh
        dd      0AA5500h
        dd      0FF7711h
        dd      0FFAAAAh
        dd      0AAAAAAh
        dd      0FFFF00h
        dd      0FFFFFFh
        dd      0555555h
        dd      00000AAh
        dd      000AA00h
        dd      000AAAAh
        dd      0AA0000h
LastColour:
        dd      0AA00AAh
        dd      -1,-1,-1,-1

.code
EndP Painter

FlashPaint Proc
        push    eax ecx edx
        call    ReadPCTimer
        mov     dx,ax
        cmp     dx,[OldTimer]
        je      FPExit
        mov     [OldTimer],dx
        xor     ecx,ecx
        mov     cl,[PaintCursor]
        mov     eax,[DWord Ptr VGBScheme+ecx*4]
        and     dl,12
        jnz     NotWhite
        xor     eax,0FFFFFFh
NotWhite:
        add     cl,10h
        call    SetVGAPalette
FPExit: pop     edx ecx eax
        ret
EndP FlashPaint

; ***********************************************************************
; *                                                                     *
; *                 Load the default painting palette                   *
; *                                                                     *
; ***********************************************************************

PaintDefault Proc
        push    eax ebx ecx edx
        mov     ebx,offset VGBScheme
        mov     al,0
PD1:    xor     ecx,ecx
        test    al,1
        jz      PD2
        mov     cl,255
PD2:    test    al,2
        jz      PD3
        mov     ch,255
PD3:    test    al,4
        jz      PD4
        or      ecx,0FF0000h
PD4:    mov     [ebx],ecx
        add     ebx,4
        inc     al
        cmp     al,16
        jb      PD1
        call    LoadColourScheme
        pop     edx ecx ebx eax
        ret
EndP PaintDefault
end
