; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *                     Global Data / Setup Data                        *
; *                                                                     *
; *     Note : this is largely based on Marcel DeKogel's Sound file     *
; *     with some simplifications.                                      *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************
.386p
.model flat
include gb.inc

ChannelSize     equ     64              ; size of channel record

chFrequency     equ     0               ; offsets for structure members
chEnvelope      equ     2               ; all are words EXCEPT chChanVoice
chFreqSweep     equ     4               ; which points to the channel voice !
chEnvSweep      equ     6
chEnvSweepPar   equ     8
chChannelOn     equ     10
chToneOff       equ     12
chKeyOn         equ     14
chVolIsNull     equ     16
chChannel       equ     18              ; the channel number
chChanVoice     equ     20
chOpAdBase      equ     24              ; OpAd[] value for this channel.
chKBF           equ     26
chLength        equ     28
.data

SoundReg db     32 dup (0)              ; image of the sound registers

Channel0 db     ChannelSize dup (0)     ; Channel storage,we're using 3
Channel1 db     ChannelSize dup (0)
Channel2 db     ChannelSize dup (0)
Channel3 db     ChannelSize dup (0)

OutputCtrl db   255                     ; Output control
       
.code
                           
; ***********************************************************************
; *                                                                     *
; *             Write to FM Chip. AL is data,BX is index                *
; *                                                                     *
; ***********************************************************************

WriteFM Proc

ifdef DEBUG                             ; debug sound routine
        push    eax
        call    DebugByte
        mov     al,'-'
        call    DebugChar
        mov     ax,bx
        call    DebugWord
        mov     al,13
        call    DebugChar
        mov     al,10
        call    DebugChar
        pop     eax
endif
        push    eax ebx ecx edx
        cmp     bh,0                    ; only 00..FF supported.
        jnz     NotOPL3                 ; we're only doing the simplest !
        mov     dx,[FMPort]             ; out index to FM Port
        xchg    al,bl                   ; data now in bl,index in al
        out     dx,al                   ; output the index.
        mov     cx,128                  ; delay for a bit.
WFMDly1:dec     cx
        jnz     WFMDly1
        in      al,dx                   ; read the fm port
        inc     dx                      ; dx now points to the data port
        mov     al,bl                   ; get the data back
        out     dx,al                   ; and output it.
        mov     cx,512                  ; delay for a bit again.
WFMDly2:dec     cx
        jnz     WFMDly2
        dec     dx                      ; dx points to the index port
        in      al,dx                   ; read it again.
NotOPL3:pop     edx ecx ebx eax
        ret
EndP WriteFM

; ***********************************************************************
; *                                                                     *
; *                     Initialise the sound card                       *
; *                                                                     *
; ***********************************************************************

SoundInitialise Proc
        mov     bx,1                    ; WriteFM(0,1)
        mov     al,0
        call    WriteFM
        mov     bx,4                    ; WriteFM(0x60,4)
        mov     al,60h
        call    WriteFM
        mov     al,80h                  ; WriteFM(0x80,4)
        call    WriteFM
        mov     dx,[FMPort]             ; tmp1 (cl) = InPort()
        in      al,dx
        mov     cl,al                   ; and save the result
        mov     bx,2                    ; WriteFM(0xFF,2)
        mov     al,0FFh
        call    WriteFM
        mov     bx,4                    ; WriteFM(0x21,4)
        mov     al,21h
        call    WriteFM
        mov     bx,4096                 ; delay for a bit.
        mov     dx,[FMPort]
SIDelay:in      al,dx                   ; keep reading the FM Port.
        dec     bx
        jnz     SIDelay
        mov     ch,al                   ; ch = read port value
        mov     bx,4                    ; WriteFM(0x60,4)
        mov     al,60h
        call    WriteFM
        mov     al,80h                  ; WriteFM(0x80,4)
        call    WriteFM

        and     cl,0E0h                 ; first one read must be zero
        jnz     ExitSoundInit
        and     ch,0E0h                 ; second one must have gone down !
        cmp     ch,0E0h
        je      ExitSoundInitFail

        mov     esi,offset Channel0     ; clear Channels 0..3
        mov     ax,0
        call    ClearChannel
        mov     esi,offset Channel1
        mov     ax,1
        call    ClearChannel
        mov     esi,offset Channel2
        mov     ax,2
        call    ClearChannel
        mov     esi,offset Channel3
        mov     ax,3
        call    ClearChannel

        mov     bx,0BDh                 ; WriteFM(0,0xBD)
        mov     al,0
        call    WriteFM
        mov     bx,8                    ; WriteFM(0,8);
        call    WriteFM
        call    SetVoices               ; define voices for channels 0..3
        jmp     ExitSoundInit
ExitSoundInitFail:
        mov     [SoundOn],0
ExitSoundInit:
        ret
EndP SoundInitialise

; ***********************************************************************
; *                                                                     *
; *                  Clear offset Channel pointed to by ESI             *
; *                                                                     *
; ***********************************************************************

ClearChannel Proc
        push    ebx
        xor     ebx,ebx                 ; clear everything out
        mov     [esi],ebx
        mov     [esi+4],ebx
        mov     [esi+8],ebx
        mov     [esi+12],ebx
        mov     [esi+16],ebx
        mov     [esi+20],ebx
        mov     [esi+24],ebx
        mov     [esi+28],ebx
        mov     [esi+chChannel],ax      ; channel id in offset
        movzx   ebx,ax
        mov     bx,[word ptr OpAd+ebx*2]; get the base operation address
        mov     [esi+chOpAdBase],bx
        mov     bx,1
        mov     [esi+chChannelOn],bx
        mov     [esi+chLength],bx
        pop     ebx
        ret
OpAd:   dw      0,1,2,8,9,0Ah,10h,11h,12h
EndP ClearChannel

; ***********************************************************************
; *                                                                     *
; *                 Set volume for channel ESI to AL                    *
; *                                                                     *
; ***********************************************************************

AdlibSetVolume Proc
        push    eax ebx ecx edx
        xor     ah,ah                           ; make AL 16 bit
        mov     [word ptr esi+chEnvelope],ax    ; ChannelEnv[] = vol

        cmp     al,0                            ; if new volume is zero..
        jnz     VolumeNotZero                   ; e.g. off
        mov     bx,1                            ; VolIsNul[] = 1
        mov     [word ptr esi+chVolIsNull],bx
        mov     bx,[word ptr esi+chToneOff]     ; if !toneoff
        cmp     bx,0
        jnz     VolumeNotZero
        call    AdlibChannelOff                 ; turn the channel off
VolumeNotZero:

        mov     bx,[word ptr esi+chVolIsNull]   ; if volume is off
        cmp     bx,0                            ; (vol is nul is nonzero)
        jz      NoVolIsNul
        xor     ebx,ebx                         ; VolIsNul[] = 0
        mov     [word ptr esi+chVolIsNull],bx
        mov     bx,[word ptr esi+chToneOff]     ; if !toneoff
        jnz     NoVolIsNul
        call    AdlibChannelOn                  ; turn the channel on.
NoVolIsNul:
                                                ; al = volume
        mov     ebx,[dword ptr esi+chChanVoice] ; bl = ChanVoice[][3]
        mov     bl,[ebx+3]
        mov     cx,[word ptr esi+chOpAdBase]    ; cx = OpAd[Channel]+0x43
        add     cx,043h
        ;
        ; this is writevolume(ax,bl,cx) inlined
        ;
        cmp     al,0                            ; if volume = 0
        jnz     WVNotZero
        mov     dl,63                           ; tl = 63 (stored in dl)
        jmp     WVContinue
WVNotZero:
        cmp     al,63                           ; top volume off at 63
        jb      NotMaxVol
        mov     al,63
NotMaxVol:

        mov     dl,bl                           ; tl = reg & 63 (stored in dl)
        and     dl,63
        mov     dh,63                           ; dh = 63-volume
        sub     dh,al

        mov     dl,dh
        jmp     WVContinue

        add     dl,dh                           ; tl = (reg&63)+(63-vol)
        cmp     dl,63                           ; top off at 63
        jb      WVContinue
        mov     dl,63
WVContinue:
        mov     al,bl                           ; al = (reg & 0xC0)
        and     al,0C0h
        or      al,dl                           ; al = (reg & 0xC0) | tl
        mov     bx,cx                           ; write to register cx
        call    WriteFM
        pop     edx ecx ebx eax
        ret
EndP AdlibSetVolume

; ***********************************************************************
; *                                                                     *
; *                     Turn channel esi on                             *
; *                                                                     *
; ***********************************************************************

AdlibChannelOn Proc
        push    eax ebx
        mov     al,[SoundOn]
        cmp     al,0
        jz      NoChannelOn
        mov     ax,[word ptr esi+chVolIsNull]   ; if volume is null no channel
        cmp     ax,0
        jnz     NoChannelOn
        mov     ax,[word ptr esi+chChannelOn]   ; if its off already....
        cmp     ax,0
        jz      NoChannelOn

        mov     al,[byte ptr esi+chKBF]         ; read kbf for the channel
        or      al,32                           ; write value is or with 32
        mov     bx,[word ptr esi+chChannel]     ; address = channel+0xB0
        add     bx,0B0h
        call    WriteFM                         ; write the value
        mov     bx,1
        mov     [word ptr esi+chKeyOn],bx       ; KeyOn[] = 1
NoChannelOn:
        pop     ebx eax
        ret
EndP AdlibChannelOn

; ***********************************************************************
; *                                                                     *
; *                     Turn channel esi off                            *
; *                                                                     *
; ***********************************************************************

AdlibChannelOff Proc
        push    eax ebx
        mov     al,[byte ptr esi+chKBF]         ; data in KBF
        mov     bx,[word ptr esi+chChannel]     ; address = channel+0xB0
        add     bx,0B0h
        call    WriteFM                         ; write to it
        xor     ax,ax                           ; KeyOn[] = 0
        mov     [word ptr esi+chKeyOn],ax
        pop     ebx eax
        ret
EndP AdlibChannelOff

; ***********************************************************************
; *                                                                     *
; *                     Set up the ADLIB voices                         *
; *                                                                     *
; ***********************************************************************

SetVoices Proc
        mov     esi,offset Channel0             ; channel 0
        mov     edi,offset SoundMelodic
        call    AdlibSetVoice
        mov     al,15
        call    AdlibSetVolume

        mov     esi,offset Channel1             ; channel 1 is the same
        mov     edi,offset SoundMelodic
        call    AdlibSetVoice
        mov     al,15
        call    AdlibSetVolume

        mov     esi,offset Channel2             ; channel 2 is wave
        mov     edi,offset SoundWave
        call    AdlibSetVoice
        mov     al,15
        call    AdlibSetVolume

        mov     esi,offset Channel3             ; channel 3 is noise
        mov     edi,offset SoundNoise
        call    AdlibSetVoice
        mov     al,15
        call    AdlibSetVolume
        ret

.data
SoundMelodic:   db      023h,021h,01eh,00h,0ffh,0ffh,08h,08h,0,0,0eh
SoundWave:      db      023h,021h,01eh,00h,0ffh,0ffh,08h,08h,0,0,0Ah
SoundNoise:     db      02eh,02eh,040h,03h,0ffh,0ffh,0fh,0fh,0,3,0eh
.code
EndP SetVoices

; ***********************************************************************
; *                                                                     *
; *            Set up a Voice ESI=Channel,EDI=Voice Data                *
; *                                                                     *
; ***********************************************************************

AdlibSetVoice Proc
        push    eax ebx ecx edx edi esi
        mov     [dword ptr esi+chChanVoice],edi ; ChannelVoice[] = voice
        mov     edx,offset SVAddress            ; edx points to address list
        mov     ch,10                           ; 11 registers to write
ASVLoop:
        mov     bx,[word ptr esi+chOpAdBase]    ; get the base address
        add     bx,[edx]                        ; make it to a register
        mov     al,[edi]                        ; byte to write
        call    WriteFM                         ; write it
        inc     edi                             ; increment voice pointer
        add     edx,2                           ; bump to next address
        dec     ch                              ; until finished
        jnz     ASVLoop
        mov     al,[edi]                        ; the one that goes at C0+channel
        mov     bx,[word ptr esi+chChannel]
        add     bx,0C0h
        call    WriteFM
        pop     esi edi edx ecx ebx eax
        ret
.data
SVAddress:
        dw      020h,023h,040h,043h,060h,063h,080h,083h,0e0h,0e3h,0c0h
.code
EndP AdlibSetVoice

; ***********************************************************************
; *                                                                     *
; *       Write to a sound register.CL contains data,BX address         *
; *                                                                     *
; ***********************************************************************

SoundWrite Proc
        push    eax ebx ecx edx esi edi ebp
        xor     bh,bh

ifdef DEBUG
        mov     ax,bx
        call    DebugWord
        mov     al,cl
        call    DebugByte
        mov     al,13
        call    DebugChar
        mov     al,10
        call    DebugChar
endif
                
        sub     bx,010h                         ; range $10-$26 to $0-$16
        cmp     bx,016h
        ja      SWError                         ; sound write error...
        movzx   ebx,bx                          ; zero extend bx=>ebx
        movzx   ecx,cl                          ; zero extend cl=>ecx
        mov     [offset SoundReg+ebx],cl        ; update the sound reg space
        call    [dword ptr SWDispatch+ebx*4]    ; call the appropriate routine
        pop     ebp edi esi edx ecx ebx eax
NoEffect:
        ret
SWError:jmp     GoDOS                           ; exit if out of range.

SWDispatch:
        dd      Mode1FreqSweep                  ; 0x10 Mode 1 Freq Sweep
        dd      Mode1Length                     ; 0x11 Mode 1 Len/Duty
        dd      Mode1EnvSweep                   ; 0x12 Mode 1 Envelope Sweep
        dd      Mode1FreqLow                    ; 0x13 Mode 1 Frequency Low
        dd      Mode1FreqHigh                   ; 0x14 Mode 1 Freq High
        dd      Mode2FreqSweep                  ; 0x15 Mode 2 Freq Sweep
        dd      Mode2Length                     ; 0x16 Mode 2 Len/Duty
        dd      Mode2EnvSweep                   ; 0x17 Mode 2 Envelope Sweep
        dd      Mode2FreqLow                    ; 0x18 Mode 2 Frequency Low
        dd      Mode2FreqHigh                   ; 0x19 Mode 2 Freq High
        dd      Mode3OnOff                      ; 0x1A Mode 3 On/Off
        dd      Mode3Length                     ; 0x1B Mode 3 Length
        dd      Mode3Level                      ; 0x1C Mode 3 Output Level
        dd      Mode3FreqLow                    ; 0x1D Mode 3 Frequency Low
        dd      Mode3FreqHigh                   ; 0x1E Mode 3 Freq High
        dd      NoEffect                        ; 0x1F (unknown ?)
        dd      Mode4Length                     ; 0x20 Mode 4 Length
        dd      Mode4Envelope                   ; 0x21 Mode 4 Envelope
        dd      Mode4Polynomial                 ; 0x22 Mode 4 Polynomial
        dd      NoEffect                        ; 0x23 Mode 4 Counter Initial
        dd      SetOutputLevel                  ; 0x24 Channel Control
        dd      SelectTerminal                  ; 0x25 Sound Output Select
        dd      SoundWrite                      ; 0x26 Sound On/Off
; ===========================================================================
;                  Write to 0x10 - Mode 1 Frequency Sweep
; ===========================================================================
Mode1FreqSweep:
        call    GetFrequencySweep               ; convert cl to freq sweep
        mov     [word ptr Channel0+chFreqSweep],cx ; and save it
        ret
; ===========================================================================
;                  Write to 0x11 - Mode 1 Length
; ===========================================================================
Mode1Length:
        and     cx,63
;       mov     [word ptr Channel0+chLength],cx
        ret
; ===========================================================================
;                  Write to 0x12 - Mode 1 Envelope Sweep
; ===========================================================================
Mode1EnvSweep:
        push    ecx                             ; save for later
        call    GetSoundVolume                  ; convert to a volume
        mov     al,cl                           ; set that channels volume
        mov     esi,offset Channel0
        call    AdlibSetVolume
        pop     ecx                             ; restore the value
        call    GetEnvelopeSweep                ; convert cl to env sweep
        mov     [word ptr Channel0+chEnvSweep],cx
        ret
; ===========================================================================
;                  Write to 0x13 - Mode 1 Frequency Low
; ===========================================================================
Mode1FreqLow:
        mov     ax,[word ptr SoundReg+03h]      ; AX = Frequency
        mov     esi,offset Channel0             ; ESI = Channel Record
        call    SetFrequency                    ; set the frequency
        ret
; ===========================================================================
;                Write to 0x14 - Mode 1 Frequency High
; ===========================================================================
Mode1FreqHigh:
        test    bl,128                          ; if high & 0x80
        jz      Mode1SetVolume
        mov     ax,[word ptr Channel0+chEnvSweep] ; read the envelope sweep
        cmp     ax,0
        jz      Mode1SetVolume                  ; if it is on reset volume
        mov     esi,offset Channel0
        mov     cl,[SoundReg+02h]
        call    GetSoundVolume                  ; convert volume
        mov     al,cl
        call    AdlibSetVolume
Mode1SetVolume:
        jmp     Mode1FreqLow                    ; redo the frequency stuff
; ===========================================================================
;                  Write to 0x15 - Mode 2 Frequency Sweep
; ===========================================================================
Mode2FreqSweep:
        call    GetFrequencySweep               ; convert cl to freq sweep
        mov     [word ptr Channel1+chFreqSweep],cx ; and save it
        ret
; ===========================================================================
;                  Write to 0x16 - Mode 2 Length
; ===========================================================================
Mode2Length:
        and     cx,63
;       mov     [word ptr Channel1+chLength],cx
        ret
; ===========================================================================
;                  Write to 0x17 - Mode 2 Envelope Sweep
; ===========================================================================
Mode2EnvSweep:
        push    ecx                             ; save for later
        call    GetSoundVolume                  ; convert to a volume
        mov     al,cl                           ; set that channels volume
        mov     esi,offset Channel1
        call    AdlibSetVolume
        pop     ecx                             ; restore the value
        call    GetEnvelopeSweep                ; convert cl to env sweep
        mov     [word ptr Channel1+chEnvSweep],cx
        ret
; ===========================================================================
;                  Write to 0x18 - Mode 2 Frequency Low
; ===========================================================================
Mode2FreqLow:
        mov     ax,[word ptr SoundReg+08h]      ; AX = Frequency
        mov     esi,offset Channel1             ; ESI = Channel Record
        call    SetFrequency                    ; set the frequency
        ret
; ===========================================================================
;                Write to 0x19 - Mode 2 Frequency High
; ===========================================================================
Mode2FreqHigh:
        test    bl,128                          ; if high & 0x80
        jz      Mode2SetVolume
        mov     ax,[word ptr Channel1+chEnvSweep] ; read the envelope sweep
        cmp     ax,0
        jz      Mode2SetVolume                  ; if it is on reset volume
        mov     esi,offset Channel1
        mov     cl,[SoundReg+07h]
        call    GetSoundVolume                  ; convert volume
        mov     al,cl
        call    AdlibSetVolume
Mode2SetVolume:
        jmp     Mode2FreqLow                    ; redo the frequency stuff
; ===========================================================================
;                       Write to 0x1A - Mode 3 On/Off
; ===========================================================================
Mode3OnOff:
        mov     al,0B0h                         ; al = 0B0h if byte & 80
        test    bl,128                          ; non-zero,0F4 otherwise
        jnz     Mode3Cont
        mov     al,0F4h
Mode3Cont:
        mov     esi,offset Channel2             ; esi = channel,al = B0/F4
        mov     bl,1                            ; bl = 1
        call    AdlibToneOn                     ; set the tone
        ret
; ===========================================================================
;                  Write to 0x1B - Mode 3 Length
; ===========================================================================
Mode3Length:
;       mov     [word ptr Channel2+chLength],cx
        ret
; ===========================================================================
;                   Write to 0x1C - Mode 3 Output Level
; ===========================================================================
Mode3Level:
        and     ebx,060h                ; value = value & 60 >> 5
        shr     ebx,5
        and     ebx,3
        mov     al,[byte ptr Mode3Volume+ebx]
        mov     esi,offset Channel2
        call    AdlibSetVolume
        ret
Mode3Volume:                                    ; volume values,mode 3
        db      0,54,52,50
; ===========================================================================
;                  Write to 0x1D - Mode 3 Frequency Low
; ===========================================================================
Mode3FreqLow:
        mov     ax,[word ptr SoundReg+0Dh]      ; AX = Frequency
        mov     esi,offset Channel2             ; ESI = Channel Record
        call    SetFrequency                    ; set the frequency
        ret
; ===========================================================================
;                Write to 0x1E - Mode 3 Frequency High
; ===========================================================================
Mode3FreqHigh:
        test    bl,128                          ; if high & 0x80
        jz      Mode3SetVolume
        mov     esi,offset Channel2
        call    AdlibChannelOff                 ; turn it off
Mode3SetVolume:
        jmp     Mode3FreqLow                    ; redo the frequency stuff
; ===========================================================================
;                  Write to 0x20 - Mode 4 Length
; ===========================================================================
Mode4Length:
        and     cx,63
        mov     [word ptr Channel3+chLength],cx
        ret
; ===========================================================================
;                 Write to 0x21 - Mode 4 Envelope
; ===========================================================================
Mode4Envelope:
        push    ecx                             ; save for later
        call    GetSoundVolume                  ; convert to a volume
        mov     al,cl                           ; set that channels volume
        mov     esi,offset Channel3
        call    AdlibSetVolume
        pop     ecx                             ; restore the value
        call    GetEnvelopeSweep                ; convert cl to env sweep
        mov     [word ptr Channel3+chEnvSweep],cx
        ret
; ===========================================================================
;              Write to 0x22 - Mode 4 Polynomial counter
; ===========================================================================
Mode4Polynomial:
        mov     esi,offset Channel3             ; set the frequency to 1100Hz
        mov     eax,1100
        call    AdlibSetFreq
        ret
; ===========================================================================
;               Write to 0x24 - Set SO1 & SO2 levels
; ===========================================================================
SetOutputLevel:
        xor     al,al
        test    bl,128
        jz      SkipSO2
        or      al,240
SkipSO2:test    bl,8
        jz      SkipSO1
        or      al,15
SkipSO1:mov     [OutputCtrl],al
        call    SelectTerminal
        ret
; ===========================================================================
;                Write to 0x25 - Select Sound Terminal
; ===========================================================================
SelectTerminal:
        mov     esi,offset Channel0             ; for each channel,check
        mov     bl,011h                         ; the two bits in 0x25
        call    SubSelect                       ; (no stereo supported)
        mov     esi,offset Channel1
        mov     bl,022h
        call    SubSelect
        mov     esi,offset Channel2
        mov     bl,044h
        call    SubSelect
        mov     esi,offset Channel3
        mov     bl,088h
        call    SubSelect
        call    Set0x26
        ret
SubSelect:
        mov     al,0F2h
        and     bl,[byte ptr SoundReg+015h]
        and     bl,[OutputCtrl]
        jz      SS2
        mov     al,0D0h
SS2:    mov     bl,1
        call    AdlibToneOn
        ret

Set0x26:mov     bl,[byte ptr SoundReg+015h]     ; figure out what's on.
;       and     bl,[OutputCtrl]
        mov     bh,bl
        shr     bh,4
        or      bl,bh
        and     bl,15
        mov     bh,[byte ptr Memory+0FF26h]     ; update $ff26
        and     bh,128
        or      bl,bh
        mov     [byte ptr Memory+0FF26h],bl
        ret
       
; ===========================================================================
;                   Write to 0x26 - Sound On/Off
; ===========================================================================
SoundControl:
        test    cl,128                  ; if high bit is clear
        mov     cl,15
        jnz     NotAllOff
        xor     cl,cl                   ; then everything is off.
NotAllOff:
        mov     esi,offset Channel0     ; turn main channels on/off
        mov     ch,1
        call    MainOnOff
        mov     esi,offset Channel1
        mov     ch,2
        call    MainOnOff
        mov     esi,offset Channel2
        mov     ch,4
        call    MainOnOff
        mov     esi,offset Channel3
        mov     ch,8
        call    MainOnOff
        call    Set0x26
        ret

MainOnOff:
        and     ch,cl
        jz      MainOff
        mov     al,0E0h                 ; turn it on
        mov     bx,[word ptr esi+chKeyOn]
        call    AdlibToneOn
        ret
MainOff:mov     al,0F1h
        mov     bl,1
        call    AdlibToneOn
        ret

EndP SoundWrite

; ***********************************************************************
; *                                                                     *
; *             Convert volume in CL to extended volume in CL           *
; *                                                                     *
; ***********************************************************************

GetSoundVolume Proc
        push    eax
        movzx   eax,cl                  ; eax = extended cl
        shr     eax,4
        xor     ch,ch
        mov     cl,[byte ptr VolumeTable+eax]
        pop     eax
        ret
VolumeTable:
        db      0,32,34,36,38,40,42,44,46,48,50,52,54,56,58,60
        ret
EndP GetSoundVolume

; ***********************************************************************
; *                                                                     *
; *     Convert envelope sweep in CL to real envelope sweep in CX       *
; *                                                                     *
; ***********************************************************************

include sweeptab.inc

GetEnvelopeSweep Proc
        and     ecx,0FFh                        ; expand cx to full 32 bit offset
        mov     cx,[word ptr EnvSweepTab+ecx*2] ; get the value
        ret
EndP GetEnvelopeSweep

; ***********************************************************************
; *                                                                     *
; *    Convert frequency sweep in CL to real frequency sweep in CX      *
; *                                                                     *
; ***********************************************************************

GetFrequencySweep Proc
        and     ecx,0FFh                        ; expand cx to full 32 bit offset
        mov     cx,[word ptr FrqSweepTab+ecx*2] ; get the value
        ret
EndP GetFrequencySweep

; ***********************************************************************
; *                                                                     *
; *                     Turn all the sound off                          *
; *                                                                     *
; ***********************************************************************

SoundOff Proc
        mov     esi,offset Channel0
        call    AdlibChannelOff
        mov     esi,offset Channel1
        call    AdlibChannelOff
        mov     esi,offset Channel2
        call    AdlibChannelOff
        mov     esi,offset Channel3
        call    AdlibChannelOff
        ret
EndP SoundOff

; ***********************************************************************
; *                                                                     *
; *             Set Frequency to upper 11 bits of AX                    *
; *                                                                     *
; ***********************************************************************

SetFrequency Proc
        push    eax
        and     eax,07FFh
        call    AdlibSetFreq
        pop     eax
        ret
EndP SetFrequency

; ***********************************************************************
; *                                                                     *
; *             Adlib ToneOn. Channel^ ESI State:AL KeyOn:BL            *
; *                                                                     *
; ***********************************************************************

AdlibToneOn Proc
        push    eax ebx ecx edx
        mov     dl,al                           ; dl = state & 15
        and     dx,15
        mov     [word ptr esi+chToneOff],dx     ; put this in the toneoff
        shr     dx,4                            ; and it with dl >> 4
        and     [word ptr esi+chToneOff],dx
        cmp     bl,0                            ; if keyon = 0
        jz      ATOOff                          ; turn it off.
        mov     bx,[word ptr esi+chToneOff]     ; if ToneOff[] then
        cmp     bx,0                            ; turn it off.
        jnz     ATOOff
        mov     bx,[word ptr esi+chLength]
        cmp     bx,0
        jz      ATOOff

        call    AdlibChannelOn                  ; turn it on & exit
        jmp     ATOExit

ATOOff: call    AdlibChannelOff                 ; turn it off & exit
ATOExit:pop     edx ecx ebx eax
        ret
EndP AdlibToneOn

; ***********************************************************************
; *                                                                     *
; *             Set the frequency of channel ESI to EAX                 *
; *                                                                     *
; ***********************************************************************

AdlibSetFreq Proc
        push    eax ebx ecx edx
        movzx   eax,ax                          ; zero extend ax
        mov     [word ptr esi+chFrequency],ax   ; save the frequency
        mov     dx,[word ptr FMFreqs+eax*4+2]   ; get freq into dx
        mov     cx,[word ptr FMFreqs+eax*4]     ; and block into cx

        mov     al,dl                           ; al = freq & FF
        mov     bx,[word ptr ESI+chChannel]     ; send it to channel+0A0
        add     bx,0A0h
        call    WriteFM

        push    dx
        and     dh,3
        and     cl,7
        shl     cl,2                            ; block = block << 2
        or      cl,dh                           ; cx = sum of the above
        xor     ch,ch
        mov     [word ptr ESI+chKBF],cx         ; save it in KBF
        pop     dx

        mov     cx,[word ptr ESI+chToneOff]
        cmp     cx,0
        jnz     ASFExit
        cmp     dx,0
        jz      ASFOff
        call    AdlibChannelOn
        jmp     ASFExit
ASFOff: call    AdlibChannelOff
ASFExit:

        pop     edx ecx ebx eax
        ret

FMFreqs:
        include FMFreqs.Inc
EndP AdlibSetFreq

; ***********************************************************************
; *                                                                     *
; *                     Sound Interrupt Routine                         *
; *                                                                     *
; ***********************************************************************

.data
EnvCount db     1                               ; Envelope sweep counter
FrqCount db     1                               ; Frequency sweep counter
RateCount db    1
.code
EnvPeriod equ   1
FreqPeriod equ  2

SoundInterruptRoutine Proc
        push    eax ebx ecx edx esi
        mov     [RateCount],5
SIRLoop:
        dec     [EnvCount]
        jnz     NotEnvSweep
        mov     [EnvCount],EnvPeriod
        mov     esi,offset Channel0
        call    ChannelEnvSweep
        mov     esi,offset Channel1
        call    ChannelEnvSweep
        mov     esi,offset Channel2
        call    ChannelEnvSweep
        mov     esi,offset Channel3
        call    ChannelEnvSweep
NotEnvSweep:
        dec     [FrqCount]
        jnz     NotFrqSweep
        mov     [FrqCount],FreqPeriod

        mov     esi,offset Channel0
        call    ChannelFrqSweep
        mov     esi,offset Channel1
        call    ChannelFrqSweep
        mov     esi,offset Channel2
        call    ChannelFrqSweep
        mov     esi,offset Channel3
        call    ChannelFrqSweep

NotFrqSweep:
;        mov     esi,offset Channel0
;        call    ChannelLength
;        mov     esi,offset Channel1
;        call    ChannelLength
;        mov     esi,offset Channel2
;        call    ChannelLength
        mov     esi,offset Channel3
        call    ChannelLength

        dec     [RateCount]
        jnz     SIRLoop
        pop     esi edx ecx ebx eax
        ret

EndP SoundInterruptRoutine

AbsAX Proc
        test    ah,128
        jz      AbsAXExit
        xor     ax,0FFFFh
        inc     ax
AbsAXExit:
        ret
EndP AbsAX

ChannelLength Proc
        mov     ax,[word ptr esi+chLength]
        cmp     ax,0
        jz      ChannelLenExit
        dec     ax
        mov     [word ptr esi+chLength],ax
        cmp     ax,0
        jnz     ChannelLenExit

        call    AdlibChannelOff
ChannelLenExit:
        ret
EndP ChannelLength

ChannelEnvSweep Proc
        ret
        mov     ax,[word ptr esi+chEnvSweep]    ; exit if no env sweep
        cmp     ax,0
        jz      ExitCES

        mov     ax,[word ptr esi+chEnvSweepPar] ; if counter > 0 dec & exit
        cmp     ax,0
        jnz     ExitCESDec

        mov     ax,[word ptr esi+chEnvSweep]    ; count = abs(envsweep)
        call    AbsAX
        mov     [word ptr esi+chEnvSweepPar],ax

        mov     bx,[word ptr esi+chEnvelope]    ; bx = envelope
        mov     ax,[word ptr esi+chEnvSweep]    ; get the sweep direction
        test    ah,0
        jnz     CESDown                         ; if negative sweep down

        add     bx,1                            ; sweep up
        cmp     bx,63                           ; as far as 63
        jb      CESSetEnvelope
        mov     bx,63
        jmp     CESSetEnvelope

CESDown:sub     bx,1                            ; sweep down while +ve
        test    bh,128
        jz      CESSetEnvelope
        xor     bx,bx
CESSetEnvelope:
        cmp     bx,[word ptr esi+chEnvelope]    ; if it hasn't changed
        je      ExitCES
        mov     [word ptr esi+chEnvelope],bx
        mov     al,bl                           ; set volume
        call    AdlibSetVolume
        jmp     ExitCES                         ; and exit

ExitCESDec:                                     ; dec & exit here
        mov     ax,[word ptr esi+chEnvSweepPar]
        dec     ax
        mov     [word ptr esi+chEnvSweepPar],ax
ExitCES:ret
EndP ChannelEnvSweep

ChannelFrqSweep Proc
        xor     eax,eax                         ; clear 32
        mov     ax,[word ptr esi+chFreqSweep]   ; check the freq sweep
        cmp     ax,0
        jz      ExitCFS
        add     ax,[word ptr esi+chFrequency]   ; add to frequency
        cmp     ax,0                            ; check range
        jnz     CFSNotZero
        xor     ax,ax
CFSNotZero:
        cmp     ax,2048
        jb      CFSNot2048
        mov     ax,2047
CFSNot2048:
        cmp     ax,[word ptr esi+chFrequency]   ; has it changed ?
        je      ExitCFS                         ; if not don't bother
        mov     [word ptr esi+chFrequency],ax   ; set the frequency
        call    AdlibSetFreq                    ; set the frequency...
ExitCFS:ret
EndP ChannelFrqSweep

end
