; ***********************************************************************
; ***********************************************************************
; *                                                                     *
; *             Write to I/O area (FF00-FFFF) and ROM Mapping           *
; *                                                                     *
; ***********************************************************************
; ***********************************************************************

.386p 
.model flat 
.code 
include Gb.Inc

.code

; ***********************************************************************
; *                                                                     *
; *        Reset the ROM & RAM masks for the current cartridge          *
; *                                                                     *
; ***********************************************************************

ResetMappingData Proc
        push    eax ebx
        movzx   ebx,[byte ptr ROMImage+0148h]   ; get the ROM Size
        mov     al,[offset ROMMaskTable+ebx]    ; get the mask
        mov     [ROMMask],al
        mov     [CurrROMPage],1                 ; current ROM page
        mov     [CurrRAMPage],255               ; current RAM page
        pop     ebx eax
        ret
ROMMaskTable:
        db      1,3,7,15,31,63,127              ; masks for ROM sizes
EndP ResetMappingData

; ***********************************************************************
; *                                                                     *
; *     Called when a write to $FFxx or ROM ($0000-$7FFF) occurs.       *
; *     CL is the data,ESI points to physical memory and BX is the      *
; *     Logical address.                                                *
; *                                                                     *
; *     Active writes not done yet                                      *
; *     Joypad ($FF00) DMA ($FF46) Serial ($FF01/2) Sound ($FF10..26)   *
; *                                                                     *
; ***********************************************************************

WritePageFF Proc
        cmp     bh,0FFh                 ; write to page 00-7F ignored
        jne     WPFFMapping             ; for now, this will be mapping later

;       cmp     bl,040h
;       je      Check40

        cmp     bl,0FFh                 ; $FFFF write bits & 0x1F
        je      WPFFWriteAnd1F

        cmp     bl,046h                 ; >$FF46 goes to memory
        ja      WPFFWrite
        je      WPFFDMAWrite            ; $FF46 goes to DMA Write

        cmp     bl,00h                  ; $FF00 is the keyboard
        jz      WPFFKeyboard
        cmp     bl,01h                  ; $FF01/2 are handled in serial.asm
        je      WPFFSerial              ; I don't really understand it
        cmp     bl,02h                  ; so its cut out for development
        je      WPFFSerial

        cmp     bl,07h                  ; $FF07 is timer control.
        je      WPFFTimerControl
        cmp     bl,0Fh                  ; $FF0F (int flags) write & 0x1F
        je      WPFFWriteAnd1F
        cmp     bl,41h                  ; $FF41 is the LCDC Status
        je      WPFFLCDStatus
        cmp     bl,44h                  ; $FF44 is the LCD line
        je      WPFFResetLCD

        cmp     bl,10h                  ; $FF10..$FF26 inclusive are sound
        jb      WPFFWrite
        cmp     bl,26h
        ja      WPFFWrite
        mov     [esi],cl                ; write the value out
        call    SoundWrite              ; write it to a sound register
        jmp     WPFFExit                ; and quit

Check40:cmp     cl,[esi]
        je      WPFFWrite
        push    ax
        mov     al,[esi]
        call    DebugByte
        mov     al,'-'
        call    DebugChar
        mov     al,cl
        call    DebugByte
        mov     al,'@'
        call    DebugChar
        mov     al,[byte ptr Memory+0FF44h]
        call    DebugByte
        mov     al,13
        call    DebugChar
        mov     al,10
        call    DebugChar
        pop     ax
        jmp     WPFFWrite

DebugWPFF:
        mov     al,cl
        call    DebugByte

WPFFWriteAnd1F:                         ; write bits 0..4 ONLY
        and     cl,01Fh                 ; used for int stuff FF0F FFFF
WPFFWrite:
        mov     [esi],cl                ; default,just write it.
WPFFExit:
        ret

; ============================================================================
;
;       write to $FF07. Bits 0..2 of this register control the
;       timer. When its off (bit 2 = 0) the cycle rate will be 32767
;       so TimedOut() will be called rarely - but it won't decrement
;       TIMECNT anyway !
;
; ============================================================================

WPFFTimerControl:
        or      cl,0F8h                 ; all other bits set (from VGBDOS)
        mov     [esi],cl                ; update the byte.
        push    eax
        push    ecx
        and     ecx,7                   ; bottom 3 bits spec timer count.
        mov     eax,[Dword Ptr WPFFTimeTab+ecx*4]
        call    SetTimerRate            ; set the timer rate.
        pop     ecx
        pop     eax
        jmp     WPFFExit                ; and exit.
WPFFTimeTab:
        dd      32767,32767,32767,32767 ; 000-011 timer almost off,no dec anyway
        dd      1024                    ; 100 4096Hz
        dd      16                      ; 101 262144Hz
        dd      64                      ; 110 65536Hz
        dd      256                     ; 111 16384Hz

; ============================================================================
;
;       write to $FF41 the LCD Status. This has two parts - bits 3..7 which
;       are user settable and bits 0..2 which come from the hardware
;       and cannot be changed. Writing only affects bits 3..7
;
; ============================================================================
WPFFLCDStatus:
        and     [byte ptr Memory+0FF41h],7      ; keep the original bits 0,1,2
        and     cl,0F8h                         ; rest from the written byte
        or      [byte ptr Memory+0FF41h],cl     ; combine the two.
        jmp     WPFFExit
; ============================================================================
;
;       write to $FF44 the Current line always resets it to zero
;
; ============================================================================
WPFFResetLCD:
        mov     [byte ptr Memory+0FF44h],0      ; any write to $FF44 resets the line
        jmp     WPFFExit                        ; LCD scanline counter
; ============================================================================
;
;       Write to $FF00 the keyboard.
;
; ============================================================================
WPFFKeyboard:
        mov     [esi],cl                        ; save the value
        call    UpdateFF00                      ; update the byte there
        mov     cl,[NewFF00Value]               ; get the new value
        mov     [esi],cl                        ; copy it into memory
        jmp     WPFFExit                        ; and quit
; ============================================================================
;
;       Write to $FF46 DMA Write
;
; ============================================================================
WPFFDMAWrite:
        push    eax ebx ecx edi esi
        mov     bh,cl                           ; bx = source address
        xor     bl,bl
        call    AtomTranslate                   ; translate into esi
        mov     edi,offset Memory+0FE00h        ; edi = $FE00 physical
        mov     ecx,160/4                       ; write 40 DWORDs
DMAWrite0:
        mov     eax,[esi]                       ; copy one dword at a time
        mov     [edi],eax                       ; in this loop.
        add     esi,4
        add     edi,4
        dec     ecx
        jnz     DMAWrite0
        pop     esi edi ecx ebx eax
        mov     [esi],cl                        ; write out the data
        jmp     WPFFExit
; ============================================================================
;
;       Write to $FF01,$FF02 Serial Control
;
; ============================================================================
WPFFSerial:
        call    SerialWrite                     ; this is done in serial.asm
        jmp     WPFFExit
; ============================================================================
;
;                 mapping - write to ROM Space 0000..7FFF
;
; ============================================================================
WPFFMapping:
        cmp     bh,20h                          ; 0000..1FFF has no effect
        jb      WPFFExit
        cmp     bh,40h                          ; 2000..3FFF maps ROM
        jb      RemapROM
        cmp     bh,60h                          ; 4000..5FFF maps RAM
        jb      RemapRAM
        jmp     WPFFExit
; ============================================================================
;
;   remap rom. And written value with the mask for the ROM, map 0 onto 1
;   check to see if its changed & if so remap the ROM space 4000-7FFF
;
; ============================================================================
RemapROM:
        push    eax                             ; save eax
        cmp     [byte ptr ROMImage+147h],0      ; type 0 ROMs can't map
        jz      ExitRemapROM
        cmp     [byte ptr ROMImage+147h],5      ; if 5 or 6 (MBC2) then
        jb      NotMBC2                         ; only writes to $2100 work
        cmp     bh,21h                          ; so if $20xx or $22xx-$3Fxx
        jne     ExitRemapROM                    ; forget it.
NotMBC2:
        xor     eax,eax                         ; get mask for number of banks
        mov     al,[ROMMask]                    ; into EAX
        and     al,cl                           ; eax = bank number & mask
        jnz     NotBank0                        ; bank number 0 is 1 really
        mov     al,1
NotBank0:
        cmp     al,[CurrROMPage]                ; do we need a switch ?
        je      ExitRemapROM                    ; not if its loaded already.
        mov     [CurrROMPage],al                ; new rom page
        shl     eax,14                          ; multiply by 16384
        add     eax,offset ROMImage             ; eax points to ROM start
        call    MapROMPage                      ; redo the mapping
ExitRemapRom:
        pop     eax                             ; restore EAX
        jmp     WPFFExit                        ; and loop back.

RemapRAM:
        push    eax
        cmp     [byte ptr ROMImage+0149h],3     ; does it have 32k RAM ?
        jne     ExitRemapRam
        mov     al,cl                           ; get the bank number
        and     al,3
        cmp     al,[CurrRAMPage]                ; do we need a switch
        je      ExitRemapRAM                    ; if not,exit now.
        mov     [CurrRAMPage],al                ; update current page
        call    MapRAMPage                      ; map the new page on.
ExitRemapRAM:
        pop     eax
        jmp     WPFFExit
EndP WritePageFF

; ***********************************************************************
; *                                                                     *
; *     Get Updated value for FF00 dependent on the keyboard byte       *
; *     (note : doesn't actually change the value in $FF00)             *
; *                                                                     *
; ***********************************************************************

UpdateFF00 Proc
        push    eax

        mov     al,[Byte Ptr Memory+0FF00h]     ; read the value
        mov     [NewFF00Value],al               ; and save it
        or      [NewFF00Value],0CFh             ; set all the flags

        test    al,020h                         ; if bit 5 is clear
        jnz     NotScanLine1                    ; scan line 1
        mov     ah,[KeyboardByte]               ; from upper bits of kbb
        xor     ah,255
        shr     ah,4
        or      ah,0F0h
        and     [NewFF00Value],ah               ; clear the bits
NotScanLine1:
        test    al,010h                         ; if bit 4 is clear
        jnz     CheckKeyInt                     ; scan line 2
        mov     ah,[KeyboardByte]               ; from lower bits of kbb
        xor     ah,255
        or      ah,0F0h
        and     [NewFF00Value],ah               ; clear the bits

CheckKeyInt:
        mov     ah,[NewFF00Value]
        mov     al,[OldPi0To3]

        cmp     ah,al
        je      NoKeyInt

        cmp     [NoInt60],0                     ; this can be disabled.
        jnz     NoKeyInt
        mov     al,010h
        call    RaiseInterrupt
NoKeyInt:

        mov     al,[NewFF00Value]
        mov     [OldPi0To3],al

        pop     eax
        ret
EndP UpdateFF00

; ***********************************************************************
; *                                                                     *
; *        Map ram page AL into A000-BFFF in Logical Space              *
; *                                                                     *
; ***********************************************************************

MapRamPage Proc
        push    eax ebx ecx
        and     eax,03h                         ; only 3 pages allowed.
        mov     [CurrRAMPage],al                ; save current RAM Page
        shl     eax,13                          ; multiply it by 8192
        add     eax,offset CartridgeRAM         ; its now a pointer

        mov     ebx,offset Mapping+0A0h*4       ; start of cart RAM map area
        mov     cl,32                           ; 32 1/4k pages to map
MapRam1:mov     [ebx],eax                       ; map it
        add     eax,256
        add     ebx,4
        dec     cl
        jnz     MapRam1
        pop     ecx ebx eax
        ret
EndP MapRamPage
end

